"""
Utilities for animdl

Scraping is often quite messy and might require
a lot of complicated algorithms. Keeping these
complicated algorithms scattered across the codebase
is not good for maintainability.
"""

from . import http_caching, http_client, optopt, powertools, searching
